@@ -1,31 +1,28 @@
 """
-Ported from a Julia package, https://github.com/Winux2k/FiniteDifferenceFormula.jl, this Python package
-provides a general finite difference formula generator and a tool for teaching/learning the finite
-difference method. It generates finite difference formulas for derivatives of various orders by using
-Taylor series expansions of a function at evenly spaced points. It also gives the truncation error of
-a formula in the big-O notation. We can use it to generate new formulas in addition to verification of
-known ones.
+Ported from a Julia package, https://github.com/Winux2k/FiniteDifferenceFormula.jl,
+this Python package provides a general finite difference formula generator and a
+tool for teaching/learning the finite difference method. It generates finite
+difference formulas for derivatives of various orders by using Taylor series
+expansions of a function at evenly spaced points. It also gives the truncation
+error of a formula in the big-O notation. We can use it to generate new formulas
+in addition to verification of known ones. By changing decimal places, we can
+also see how rounding errors may affect a result.
 
-We may play with this package when teaching/learning numerical computing, especially the finite
-difference method, and explore the distribution and symmetry in the coefficients of the formulas.
-By changing decimal places, we can also see how rounding errors may affect a result.
+Beware, though formulas are mathematically correct, they may not be numerically
+useful. This is true especially when we derive formulas for a derivative of higher
+order. For example, run compute(9,range(-5, 6)), provided by this package, to
+generate a 10-point central formula for the 9-th derivative. The formula is
+mathematically correct, but it can hardly be put into use for numerical computing
+without, if possible, rewriting it in a special way. Similarly, the more points
+are used, the more precise a formula is mathematically. However, due to rounding
+errors, this may not be true numerically.
 
-Beware, though formulas are mathematically correct, they may not be numerically useful. This is true
-especially when we derive formulas for a derivative of higher order. For example, run
-compute(9,range(-5, 6)), provided by this package, to generate a 10-point central formula for the
-9-th derivative. The formula is mathematically correct, but it can hardly be put into use for
-numerical computing without, if possible, rewriting it in a special way. Similarly, the more points
-are used, the more precise a formula is mathematically. However, due to rounding errors, this may
-not be true numerically.
+The package exports a class, FDFormula, fd (an object of the class), and member
+functions, activatepythonfunction, compute, decimalplaces, find, findbackward,
+findforward, formula, formulas, taylor, taylorcoefs, tcoefs, truncationerror,
+verifyformula.
 
-To run the code, you need the Python programming language (https://python.org/), a wonderful and
-amazing computing platform.
-
-The package exports a class, FDFormula, fd (an object of the class), and member functions,
-activatepythonfunction, compute, decimalplaces, find, findbackward, findforward, formula, formulas,
-taylor, taylorcoefs, tcoefs, truncationerror, verifyformula.
-
-See also https://github.com/Winux2k/FiniteDifferenceFormula.py/blob/main/README.md
+See also https://github.com/fdformula/FiniteDifferenceFormula.py/blob/main/README.md.
 """
 #-------------------------------------------------------------------------------
 # Name:        FiniteDifferenceFormula
@@ -202,7 +199,6 @@ class FDFormula:
             if num_of_nonzero_terms == 0:
                 break
         print(" + ...")
-        return
     # end of _print_taylor
 
     def _dashline(self, ):
@@ -276,7 +272,7 @@ class FDFormula:
         """
         points = self._validate_input(n, points, printformulaq)
         if points == []:
-            return
+            return None
         return self._compute(n, points, printformulaq)
     # end of compute
 
@@ -298,7 +294,7 @@ class FDFormula:
         """
         points = self._validate_input(n, points, printformulaq)
         if points == []:
-            return
+            return None
         result = self._compute(n, points, printformulaq)
         while result == None and len(points) > n + 1:   # failed
             if self._range_inputq:
@@ -321,7 +317,7 @@ class FDFormula:
     def _findforward(self, n, points, printformulaq = False, forwardq:bool = True):
         points = self._validate_input(n, points, printformulaq)
         if points == []:
-            return
+            return None
         result = self._compute(n, points, printformulaq)
         while result == None and len(points) > n + 1:   # failed
             #points = forwardq ? points[2 : end] : points[1 : end - 1]
@@ -415,7 +411,7 @@ class FDFormula:
         #     th = n == 1 ? "st" : (n == 2 ? "nd" : (n == 3 ? "rd" : "th"))
         #     println("$pts is invalid because at least $(n + 1) points are ",
         #             "needed for the $n$th derivative.")
-        #     return
+        #     return None
         # end
         #
         # setup a linear system Ax = B first
@@ -596,7 +592,6 @@ class FDFormula:
         self._computedq = True
 
         self._test_formula_validity()
-        return
     # end of loadcomputingresults
 
     # input: A, n x (n + 1) "matrix" of Fraction numbers
@@ -702,7 +697,6 @@ class FDFormula:
         #    k[1]*coefs[1][j] + k[2]*coefs[2][j] + ... + k[len]*coefs[len][j] = 0
         # where j = 1:n
         n, k   = self._data.n, self._data.k
-        coefs  = self._data.coefs
         points = self._data.points
         length = len(points)
 
@@ -949,8 +943,6 @@ class FDFormula:
                       self._python_decimal_func_expr, "\n", sep = '')
         else:
             print("\nNo formula can be found.")
-
-        return
     # end of formula
 
     def truncationerror(self):
@@ -1037,11 +1029,11 @@ class FDFormula:
         if (not isinstance(n, int)) or n <= 0:
             print("Error: invalid first argument, ", n, ". A positive ",
                   "integer is expected.", sep = '')
-            return
+            return None
         if len(points) == 1:
             print("Error: invalid input, points = ", points, ". A list of ",
                   "two or more points is expected.", sep = '')
-            return
+            return None
         all_intq = True
         for i in points:
             if not isinstance(i, int):
@@ -1050,16 +1042,16 @@ class FDFormula:
         if not all_intq:
             print("Error: invalid input, ", points, ". Integers are expected.",
                   sep = '')
-            return
+            return None
 
         if isinstance(points, tuple):
             print("Invalid input, ", points,
                   ". A list like [1, 2, ...] is expected.", sep = '')
-            return
+            return None
         if isinstance(k, tuple):
             print("Invalid input, ", k, ". A list like [1, 2, ...] is expected",
                   sep = '')
-            return
+            return None
 
         oldlen = len(points)
         points = sorted(set(points))
@@ -1070,11 +1062,11 @@ class FDFormula:
         #if n <= len
         #    println("Error: at least $(n+1) points are needed for the $(_nth(n))",
         #            " derivative.")
-        #    return
+        #    return None
         #end
         if length != len(k):
             print("Error: The number of points != the number of coefficients.");
-            return
+            return None
 
         self._initialization()      # needed b/c it's like computing a new formula
         input_points = self._format_of_points(points)
@@ -1091,8 +1083,8 @@ class FDFormula:
             if not isinstance(m, Fraction):
                 m = Fraction(m).limit_denominator()   # rationalize
             for i in range(length):
-                k[i] *= m.den
-            m = m.num
+                k[i] *= m.denominator
+            m = m.numerator
             rewrittenq = True
         if m == 0:
             if rewrittenq:
@@ -1100,7 +1092,7 @@ class FDFormula:
                       sep = '')
             print("Error: invalid input, the last argument m = 0. ",
                   "It can't be zero.")
-            return
+            return None
 
         # "normalize" k[:] so that each element is integer
         all_intq = True
@@ -1113,7 +1105,7 @@ class FDFormula:
             if isinstance(k[i], Fraction):
                 if k[i].denominator == 1:
                     continue
-                m         *= k[i].denominator
+                m *= k[i].denominator
                 rewrittenq = True
                 d = k[i].denominator
                 for j in range(length):
@@ -1123,7 +1115,7 @@ class FDFormula:
                 k[i] = round(k[i])
 
             # print k[:] nicely
-            ks = "[%d" % k[1]
+            ks = "[%d" % k[0]
             for i in range(1, length):
                 ks += ", %d" % k[i]
             ks += "]"
@@ -1141,7 +1133,7 @@ class FDFormula:
         # Taylor series of the linear combination
         # k[1]*f(x[i+points[1]]) + k[2]*f(x[i+points[2]]) + ...
         self._lcombination_coefs = [Fraction(0,1)] * max_num_of_terms
-        for i in range(1, length):
+        for i in range(length):
             if k[i] == 0:
                 continue
             for j in range(max_num_of_terms):
@@ -1185,7 +1177,7 @@ class FDFormula:
                       "fd.activatepythonfunction() to activate the new",
                       "Python function(s).")
                 return result
-        return
+        return None
     # end of verifyformula
 
     def _printexampleresult(self, suffix, exact):
@@ -1258,15 +1250,13 @@ class FDFormula:
         # 250, a sepcial value for communication w/ 'verifyformula'
         if not (external_dataq and self._formula_status == 250):
             print("\nCall fd.formula() to view the very definition.")
-
-        return
     # end of activatepythonfunction
 
     def tcoefs(self, j, n = 10):
         """
         Same as taylorcoefs(j, n).
         """
-        self.taylorcoefs(j, n)
+        return self.taylorcoefs(j, n) # v0.6.9 add 'return'
 
     def taylorcoefs(self, j, n = 10):
         """
@@ -1281,13 +1271,13 @@ class FDFormula:
         """
         if n < 1:
             print("n = %d? It is expected to be an positive integer." % n)
-            return
+            return None
 
         if isinstance(j, int):
             return self._taylor_coefs(j, n)
         else:
             print("Invalid input, ", j, ". An integer is expected.")
-            return
+            return None
     # end of taylorcoefs
 
     # print readable Taylor series expansion of f(x[i + j]) about x[i]
@@ -1301,7 +1291,6 @@ class FDFormula:
             js = "%d" % j
         print("f(x[i", js, "]) = ", sep = '', end = '')
         self._print_taylor(coefs, n)
-        return
     # end of _printtaylor1
 
     # print readable Taylor series of a function/expression about x[i]. e.g.,
@@ -1310,9 +1299,6 @@ class FDFormula:
     def _printtaylor2(self, coefs, n = 10):
         self._print_taylor(coefs, n)
 
-        return
-    # end of _printtaylor1
-
     # awkard implementation: Julia's multiple dispatchment is super...
     #
     # input: points_k can be an integer, a list of coefficients of a Taylor
@@ -1410,7 +1396,6 @@ class FDFormula:
             for j in range(max_num_of_terms):
                 coefs[j] += k[i] * self._taylor_coefs(points[i], max_num_of_terms)[j]
         self._print_taylor(coefs, n)
-        return
     # end of taylor
 
     # return the number of points actually used in a formula
@@ -1440,13 +1425,13 @@ class FDFormula:
         fd.formulas(3, 4, 11)              # the 3rd derivative
         """
         if min_num_of_points < 2:
-            print("Error: Invalid input, min_num_of_points =",
-                  min_num_of_points, ". It must be greater than 1.")
+            print("Error: Invalid input, min_num_of_points = ",
+                  min_num_of_points, ". It must be at least 2.", sep = '')
             return
         elif max_num_of_points < min_num_of_points:
-            print("Error: Invalid input, max_num_of_points =",
-                  max_num_of_points, ". It must be greater than",
-                  min_num_of_points)
+            print("Error: Invalid input, max_num_of_points = ",
+                  max_num_of_points, ". It must be at least ",
+                  min_num_of_points, ".", sep = '')
             return
 
         if isinstance(orders, int) and orders >= 1:
@@ -1456,13 +1441,13 @@ class FDFormula:
         elif isinstance(orders, list):
             for i in orders:
                 if not (isinstance(i, int) and i >= 1):
-                    print("Error: Invalid input, orders =", orders, ".",
-                          "A list of positive integers are expected,")
+                    print("Error: Invalid input, orders = ", orders, ". ",
+                          "A list of positive integers are expected.", sep = '')
                     return
         else:
-            print("Error: Invalid input, orders =", orders, ".",
-                  "An positive integer or a list of positive integers",
-                  "are expected,")
+            print("Error: Invalid input, orders = ", orders, ". ",
+                  "An positive integer or a list of positive integers ",
+                  "are expected.", sep = '')
             return
 
         oldlen = len(orders)
@@ -1471,7 +1456,6 @@ class FDFormula:
             print("Your input: formulas(", orders, ", ", min_num_of_points,
                   ", ", max_num_of_points, ")", sep = '')
 
-        half = round(max_num_of_points / 2)
         for n in orders:
             # forward schemes
             start = max(n + 1, min_num_of_points)
@@ -1491,8 +1475,9 @@ class FDFormula:
                     self._print_bigo_formula(self._data, self._bigO)
 
             # central schemes
-            for num_of_points in range(math.floor(max(n, min_num_of_points) / 2), \
-                                       math.ceil(max_num_of_points / 2) + 1):
+            start = math.floor(max(n, min_num_of_points) / 2)
+            stop  = math.ceil(max_num_of_points / 2) + 1
+            for num_of_points in range(start, stop):
                 length = 2 * num_of_points + 1
                 if n >= length:
                     continue
@@ -1502,6 +1487,8 @@ class FDFormula:
                     if x <= max_num_of_points:
                         print(x, "-point central finite difference formula:", sep = '')
                         self._print_bigo_formula(self._data, self._bigO)
+                    #else:
+                    #   self._initialization() # v0.6.9
     # end of formulas
 
 # end of class FDFormula:
